@@ -91,3 +91,59 @@ mutate_cond <- function(.data, condition, ..., envir = parent.frame()) {
     dplyr::mutate(...)
   .data
 }
+
+
+FrobeniusNorm <- function(M, P, E) {
+  sqrt(sum((M - P %*% E)^2))
+}
+
+is.wholenumber <- function(x, tol = .Machine$double.eps) {
+  abs(x - round(x)) < tol
+}
+
+
+# Signals ----------------------------------------------------------------
+
+send_info <- function(...) {
+  cli::cli_alert_info(
+    cli::style_bold(get_timestamp(), cli::col_blue(...))
+  )
+}
+
+send_success <- function(...) {
+  cli::cli_alert_success(
+    cli::style_bold(get_timestamp(), cli::col_green(...))
+  )
+}
+
+send_warning <- function(...) {
+  cli::cli_alert_warning(
+    cli::style_bold(get_timestamp(), cli::col_yellow(...))
+  )
+}
+
+send_error <- function(...) {
+  cli::cli_alert_danger(
+    cli::style_bold(get_timestamp(), cli::col_red(...))
+  )
+}
+
+send_elapsed_time <- function(timer) {
+  timer <- Sys.time() - timer
+  send_info(round(timer, 3), " ", attr(timer, "units"), " elapsed.")
+}
+
+get_timestamp <- function() {
+  paste0("[", Sys.time(), "]: ")
+}
+
+# https://stackoverflow.com/questions/14469522/stop-an-r-program-without-error
+send_stop <- function(...) {
+  args <- list(...)
+  if (length(args) != 0) {
+    send_error(...)
+  }
+  opt <- options(show.error.messages = FALSE)
+  on.exit(options(opt))
+  stop()
+}
